# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import sys
import types
from contextlib import contextmanager

import six

try:
    import numpy
except ImportError:
    numpy = None


ignore_vars = [
    "__ipy_scope__",  # Added by nbtutor to the calling frame globals
    "__name__",
    "__builtin__",
    "__builtins__",
    "__module__",
    "__qualname__",
    "__doc__",
    "__dict__",
    "__package__",
    "__weakref__",
]

float_types = [
    float,
]

primitive_types = [
    int,
    float,
    bool,
    type(None),
    complex,
]

sequence_types = [
    set,
    tuple,
    list,
]

array_types = [
]

key_value_types = [
    dict,
]

primitive_types.extend(list(six.string_types))

if numpy is not None:
    new_types = []
    np_types = list(set(numpy.typeDict.values()))
    np_type_names = [t.__name__ for t in np_types]
    for _type in primitive_types:
        for i, name in enumerate(np_type_names):
            if _type.__name__ in name or isinstance(np_types[i], (_type, )):
                new_types.append(np_types[i])

    for i, name in enumerate(np_type_names):
        if 'float' in name or isinstance(np_types[i], (float, )):
            float_types.append(np_types[i])

    primitive_types.extend(new_types)
    array_types.append(numpy.ndarray)

float_types = tuple(float_types)
primitive_types = tuple(primitive_types)
sequence_types = tuple(sequence_types)
array_types = tuple(array_types)
key_value_types = tuple(key_value_types)


def filter_dict(d, exclude):
    """Return a new dict with specified keys excluded from the origional dict

    Args:
        d (dict): origional dict
        exclude (list): The keys that are excluded
    """
    ret = {}
    for key, value in d.items():
        if key not in exclude:
            ret.update({key: value})
    return ret


@contextmanager
def redirect_stdout(new_stdout):
    """Redirect the stdout

    Args:
        new_stdout (io.StringIO): New stdout to use instead
    """
    old_stdout, sys.stdout = sys.stdout, new_stdout
    try:
        yield None
    finally:
        sys.stdout = old_stdout


def format(obj, options):
    """Return a string representation of the Python object

    Args:
        obj: The Python object
        options: Format options
    """
    formatters = {
        float_types: lambda x: '{:.{}g}'.format(x, options.digits),
    }
    for _types, fmtr in formatters.items():
        if isinstance(obj, _types):
            return fmtr(obj)
    try:
        if six.PY2 and isinstance(obj, six.string_types):
            return str(obj.encode('utf-8'))
        return str(obj)
    except:
        return 'OBJECT'


def get_type_info(obj):
    """Get type information for a Python object

    Args:
        obj: The Python object

    Returns:
        tuple: (object type "catagory", object type name)
    """
    if isinstance(obj, primitive_types):
        return ('primitive', type(obj).__name__)
    if isinstance(obj, sequence_types):
        return ('sequence', type(obj).__name__)
    if isinstance(obj, array_types):
        return ('array', type(obj).__name__)
    if isinstance(obj, key_value_types):
        return ('key-value', type(obj).__name__)
    if isinstance(obj, types.ModuleType):
        return ('module', type(obj).__name__)
    if isinstance(obj, (types.FunctionType, types.MethodType)):
        return ('function', type(obj).__name__)
    if isinstance(obj, type):
        if hasattr(obj, '__dict__'):
            return ('class', obj.__name__)
    if isinstance(type(obj), type):
        if hasattr(obj, '__dict__'):
            cls_name = type(obj).__name__
            if cls_name == 'classobj':
                cls_name = obj.__name__
                return ('class', '{}'.format(cls_name))
            if cls_name == 'instance':
                cls_name = obj.__class__.__name__
            return ('instance', '{} instance'.format(cls_name))

    return ('unknown', type(obj).__name__)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

from bdb import Bdb as StdBdb
from bdb import BdbQuit
from io import StringIO

from .history import Heap, StackFrames, TraceHistory
from .utils import filter_dict, ignore_vars, redirect_stdout


class Bdb(StdBdb, object):
    """
    Args:
        ipy_shell: IPython interactive shell
        options: Cell magic options

    Attributes:
        ipy_shell: IPython interactive shell
        options: Cell magic options
        code_error (bool): True if the Cell code executed successfully, else False
        stdout (io.StringIO): The stdout used instead on the IPython stdout
        trace_history (:class:`~.history.TraceHistory`): The
            line-by-line trace history data
    """

    def __init__(self, ipy_shell, options):
        super(Bdb, self).__init__()
        self.ipy_shell = ipy_shell
        self.options = options
        self.code_error = False
        self.stdout = StringIO()
        self.trace_history = TraceHistory(options)

    def run_cell(self, cell):
        """Run the Cell code using the IPython globals and locals

        Args:
            cell (str): Python code to be executed
        """
        globals = self.ipy_shell.user_global_ns
        locals = self.ipy_shell.user_ns
        globals.update({
            "__ipy_scope__": None,
        })
        try:
            with redirect_stdout(self.stdout):
                self.run(cell, globals, locals)
        except:
            self.code_error = True
            if self.options.debug:
                raise BdbQuit
        finally:
            self.finalize()

    def user_call(self, frame, argument_list):
        """This method is called when there is the remote possibility that we
        ever need to stop in this function."""
        self.get_stack_data(frame, None, 'call')

    def user_line(self, frame):
        """This function is called when we stop or break at this line."""
        self.get_stack_data(frame, None, 'step_line')

    def user_return(self, frame, return_value):
        """This function is called when a return trap is set here."""
        self.get_stack_data(frame, None, 'return')

    def user_exception(self, frame, exc_info):
        """This function is called if an exception occurs,
        but only if we are to stop at or just below this level."""
        pass

    def is_notebook_frame(self, frame):
        """Return True if the current frame belongs to the notebook, else
        False"""
        return "__ipy_scope__" in frame.f_globals.keys()

    def is_other_cell_frame(self, frame):
        """Return True if the current frame belongs to the Cell, else
        False"""
        return frame.f_code.co_filename.startswith("<ipython-input-")

    def get_stack_data(self, frame, traceback, event_type):
        """Get the stack frames data at each of the hooks above (Ie. for each
        line of the Python code)"""
        heap_data = Heap(self.options)
        stack_data = StackFrames(self.options)
        stack_frames, cur_frame_ind = self.get_stack(frame, traceback)

        for frame_ind, (frame, lineno) in enumerate(stack_frames):
            skip_this_stack = False

            # Skip the self.run calling frame (first frame)
            if frame_ind == 0:
                continue

            # Skip stack after a certain stack frame depth
            if len(stack_data) > self.options.depth:
                skip_this_stack = True
                break

            # Skip stack when frames dont belong to the current notebook or
            # current cell, I.e. frames in another global scope altogether
            # or frames in other cells
            if (not self.is_notebook_frame(frame) or
                    self.is_other_cell_frame(frame)):
                if not self.options.step_all:
                    skip_this_stack = True
                    break
                lineno = 0  # So line markers dont display for these frames
            else:
                lineno += 1  # Because cell magic is actually line 1

            # Filter out ignored names from the frame locals
            user_locals = filter_dict(
                frame.f_locals,
                ignore_vars + list(self.ipy_shell.user_ns_hidden.keys())
            )

            # Add frame and heap data
            stack_data.add(frame, lineno, event_type, user_locals)
            heap_data.add(user_locals)

        if not skip_this_stack and not stack_data.is_empty():
            self.trace_history.append(
                    stack_data,
                    heap_data,
                    self.stdout.getvalue()
            )

    def finalize(self):
        """Finalize the trace history data after execution is complete"""
        self.trace_history.sort_frame_locals()



#!/usr/bin/env python
# encoding: utf-8
import os

from nbtutor._version import __version__


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]

if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'nbtutor'
year = '2016'
author = 'Logan Page'
copyright = '{0}, {1}'.format(year, author)
version = __version__
release = __version__

language = None
todo_include_todos = True
pygments_style = 'sphinx'
templates_path = ['_templates']

napoleon_google_docstring = True
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_ivar = True
napoleon_use_rtype = True
napoleon_use_param = True

htmlhelp_basename = 'nbtutordoc'

html_theme = 'alabaster'
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

latex_elements = {}
latex_documents = [
    (master_doc, 'nbtutor.tex', 'nbtutor Documentation',
     'Logan Page', 'manual'),
]
man_pages = [(master_doc, 'nbtutor', 'nbtutor Documentation', [author], 1)]
texinfo_documents = [
    (master_doc, 'nbtutor', 'nbtutor Documentation', author, 'nbtutor',
     'One line description of project.', 'Miscellaneous'),
]
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.5', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
}


def run_apidoc(_):
    from sphinx.apidoc import main

    src_dir = os.path.abspath('nbtutor')
    if os.getcwd().endswith('docs'):  # hack to get rtfd to work
        src_dir = os.path.abspath('../nbtutor')

    docs_dir = os.path.abspath(os.path.dirname(__file__))
    api_dir = os.path.join(docs_dir, 'api')
    main(['-f', '-P', '-e', '-o', api_dir, src_dir, '--force'])


def setup(app):
    app.connect('builder-inited', run_apidoc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import

import os
import sys

from subprocess import Popen

from tools import _env


def main(**opts):
    args = [
        _env.node_bin("browserify"),
        "--standalone", "nbtutor-notebook",
    ] + _env.extension + _env.external(
        "d3",
        "uuid",
        "jquery",
        "jsplumb",
        "nbtutor-deps",
        "base/js/events",
        "base/js/dialog",
        "base/js/namespace",
        "notebook/js/celltoolbar",
    ) + _env.transform + [
        "--outfile", os.path.join(_env.DIST, "js", "nbtutor.notebook.min.js"),
        os.path.join(_env.SRC, "es6", "notebook", "notebook.es6")
    ] + opts.get("browserify", [])
    return Popen(args, shell=_env.IS_WIN).wait()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup


# should be loaded below
__version__ = None

with open('nbtutor/_version.py') as version:
    exec(version.read())

with open('./README.rst') as readme:
    long_description = readme.read()


setup(
    name="nbtutor",
    version=__version__,
    description="Visualize Python code execution in Jupyter Notebook cells",
    long_description=long_description,
    author="Logan Page",
    author_email="page.lg@gmail.com",
    license="BSD 3-Clause",
    url='https://github.com/lgpage/nbtutor',
    keywords="ipython jupyter notebook python tutor visualize code exection",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: IPython",
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License"
    ],
    packages=["nbtutor"],
    install_requires=[
        "notebook",
    ],
    tests_require=[
        "check-manifest",
        "coverage",
        "pytest",
        "pytest-cov",
        "requests",
        "sphinxcontrib-spelling",
    ],
    include_package_data=True,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import

import os
import shutil
import sys

from tools import _env


def main(**opts):
    to_clean = [
        os.path.join(_env.DIST, "js"),
        os.path.join(_env.DIST, "css")
    ]
    for path in to_clean:
        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import

import os
import shutil
import sys

from tools import _env


def main(**opts):
    shutil.copyfile(
        os.path.join(_env.SRC, "js", "index.js"),
        os.path.join(_env.DIST, "js", "nbtutor.min.js")
    )
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import

import os
import sys

from subprocess import Popen

from tools import _env


def main(**opts):
    args = [
        _env.node_bin("r.js{}".format(".cmd" if _env.IS_WIN else "")),
        "-o", os.path.join(_env.SRC, "js", "build.js"),
    ] + opts.get("requirejs", [])
    return Popen(args, shell=_env.IS_WIN).wait()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
import os
import platform


PKG = "nbtutor"

REPO_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
PKG_ROOT = os.path.join(REPO_ROOT, PKG)
SRC = os.path.join(REPO_ROOT, "src")
DIST = os.path.join(PKG_ROOT, "static", PKG)

IS_WIN = "Windows" in platform.system()


def node_bin(*it):
    return os.path.join(REPO_ROOT, "node_modules", ".bin", *it)


def external(*modules):
    # browserify --external
    return sum([["--external", m] for m in modules], [])

transform = [
    "--transform", "[", "babelify", "--sourceMapRelative", ".", "]",
]

extension = ["--extension", "es6"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import, print_function

import sys

from importlib import import_module
from multiprocessing import Pool, cpu_count

from tools import clean
from tools import requirejs


try:
    CPU_COUNT = cpu_count()
except NotImplementedError:
    CPU_COUNT = 1


def _run(mod_opt, *args, **opts):
    print("started {0} with {1}".format(*mod_opt))
    task = import_module("tools.{}".format(mod_opt[0]))
    task.main(**mod_opt[1])
    print("...completed {}".format(mod_opt[0]))
    return 0


def main(**opts):
    tasks = [
        "less",
        "deps",
        "index",
        "notebook",
    ]

    clean.main()
    pool = Pool(processes=CPU_COUNT)
    pool.map(_run, zip(tasks, [opts] * len(tasks)))
    requirejs.main(**opts)
    return 0


if __name__ == "__main__":
    opts = {}
    args = sys.argv[1:]
    if "release" in args:
        opts.update(
            browserify=["-g", "uglifyify"],
        )
    if "dev" in args:
        opts.update(
            browserify=["--debug"],
            less=["--source-map-map-inline"],
        )
    sys.exit(main(**opts))



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from os.path import join

from ._version import __version__  # flake8: noqa
from ._version import __version_info__  # flake8: noqa


# IPython Extension points
def load_ipython_extension(ip):
    from .ipython.magic import NbtutorMagics  # pylint: disable=cyclic-import
    ip.register_magics(NbtutorMagics)


# Jupyter Extension points
def _jupyter_nbextension_paths():
    return [dict(
        section="notebook",
        src=join("static", "nbtutor"),
        dest="nbtutor",
        require="nbtutor/js/nbtutor.min")]




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import json

from ipykernel import jsonutil

from .utils import filter_dict, format, get_type_info, ignore_vars


class StackFrames(object):

    def __init__(self, options):
        self.options = options
        self.data = list()

    def clear(self):
        self.__init__(self.options)

    def _add_frame(self, frame, lineno, event):
        name = frame.f_code.co_name
        name = "Global frame" if name == "<module>" else name
        self.data.append(dict({
            "id": id(frame),
            "name": name,
            "lineno": lineno,
            "event": event,
            "options": {
                "inline": self.options.inline,
            },
            "vars": list(),
        }))

    def _add_frame_locals(self, filtered_locals):
        frame_data = self.data[-1]
        for name, obj in filtered_locals.items():
            frame_data["vars"].append(dict({
                "id": id(obj),
                "name": name,
            }))

    def add(self, frame, lineno, event, filtered_locals):
        self._add_frame(frame, lineno, event)
        self._add_frame_locals(filtered_locals)

    def is_empty(self):
        return len(self) == 0

    def dumps(self):
        return json.dumps(self.data)

    def clean(self):
        return jsonutil.json_clean(self.data)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)


class Heap(object):

    def __init__(self, options):
        self.options = options
        self.data = list()
        self._new_ids = list()
        self._tmp_refs = dict()

    def clear(self):
        self.__init__(self.options)

    @property
    def object_ids(self):
        ret = list()
        for obj in self.data:
            ret.append(obj['id'])
        return ret

    def has_object(self, obj_id):
        return obj_id in self.object_ids

    def _add_object(self, obj, type_info, **kwargs):
        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        type_catagory, type_name = type_info
        position = kwargs.get('position', 'right')
        if type_catagory == 'primitive':
            position = 'right' if not self.options.nolies else position

        self._new_ids.append(obj_id)
        self.data.append(dict({
            "id": obj_id,
            "type": type_name,
            "catagory": type_catagory,
            "options": {
                "position": position,
            },
            "value": format(obj, self.options),
        }))

    def _add_array_data_object(self, obj, type_info, **kwargs):
        type_catagory, type_name = type_info
        if type_name != 'ndarray':
            self._add_object(obj, type_info, **kwargs)
            return  # XXX: Only numpy ndarrays supported for now

        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        max_reached = False
        data_values = list()
        self._new_ids.append(obj_id)
        for ind, val in enumerate(obj.flatten()):
            if ind >= self.options.max_size:
                max_reached = True
                break
            data_values.append(format(val, self.options))

        self.data.append(dict({
            "id": obj_id,
            "type": '{} array ({} x {} bytes)'.format(
                obj.dtype,
                obj.size,
                obj.itemsize
            ),
            "catagory": type_catagory,
            "options": {
                "position": kwargs.get('position', 'right'),
                "ellipsis": max_reached,
            },
            "values": data_values,
        }))

    def _add_array_object(self, obj, type_info, **kwargs):
        type_catagory, type_name = type_info
        if type_name != 'ndarray':
            self._add_object(obj, type_info, **kwargs)
            return  # XXX: Only numpy ndarrays supported for now

        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        # Simply point to flat (inline) array data
        if not self.options.expand_arrays and obj.ndim == 1:
            self._add_array_data_object(obj, type_info, **kwargs)
            self.data[-1]['type'] = '{} ({})'.format(type_name, obj.dtype)
            return

        # Get the ndarray instance data
        shape = obj.shape  # actually a class property
        strides = obj.strides  # actually a class property

        # store ref so objects are not garbage collected until the end of
        # execution, so that the id cannot, on the off chance, be re-used
        self._tmp_refs.update({(obj_id, 'shape'): shape})
        self._tmp_refs.update({(obj_id, 'strides'): strides})

        # Get underlying data object
        base_obj = obj
        while base_obj.base is not None:
            base_obj = base_obj.base

        # Array pointer offset (bytes)
        offset = (
            obj.__array_interface__['data'][0] -
            base_obj.__array_interface__['data'][0]
        )

        # If obj.base is None
        if base_obj is obj:
            base_obj = base_obj.copy()  # To get a unique id
            self._tmp_refs.update({(obj_id, 'base'): base_obj})

        # If base_obj.base is None
        if (id(base_obj), 'base') in self._tmp_refs.keys():
            base_obj = self._tmp_refs[(id(base_obj), 'base')]

        # Add the ndarray instance data
        self._add_array_data_object(base_obj, type_info, **kwargs)
        self._new_ids.append(obj_id)
        self._add(shape)
        self._add(strides)

        class_vars = dict({
            'data': base_obj,
            'shape': shape,
            'strides': strides,
        })

        type_name = '{} ({} bytes offset)'.format(type_name, offset)
        type_info = ('key-value', type_name)
        self._add_key_value_object(class_vars, type_info, **kwargs)
        self.data[-1]["id"] = obj_id
        self._new_ids.remove(id(class_vars))

    def _add_sequence_object(self, obj, type_info, **kwargs):
        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        data_values = list()
        max_reached = False
        self._new_ids.append(obj_id)
        for ind, val in enumerate(obj):
            if ind >= self.options.max_size:
                max_reached = True
                break
            self._add(val)
            data_values.append(dict({
                "id": id(val),
            }))

        type_catagory, type_name = type_info
        self.data.append(dict({
            "id": obj_id,
            "type": type_name,
            "catagory": type_catagory,
            "options": {
                "inline": self.options.inline,
                "position": kwargs.get('position', 'center'),
                "ellipsis": max_reached,
            },
            "values": data_values,
        }))

    def _add_class_object(self, obj, type_info, **kwargs):
        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        self._new_ids.append(obj_id)
        type_info = ('key-value', type_info[-1])
        class_vars = filter_dict(obj.__dict__, ignore_vars)
        self._add_key_value_object(class_vars, type_info, **kwargs)
        self.data[-1]["id"] = obj_id
        self._new_ids.remove(id(class_vars))

    def _add_key_value_object(self, obj, type_info, **kwargs):
        obj_id = id(obj)
        if self.has_object(obj_id) or obj_id in self._new_ids:
            return

        obj_keys = obj.keys()
        try:
            obj_keys = sorted(obj_keys, key=lambda x: str(x))
        except:
            pass

        data_values = list()
        max_reached = False
        self._new_ids.append(obj_id)
        for ind, key in enumerate(obj_keys):
            if ind >= self.options.max_size:
                max_reached = True
                break
            value = obj[key]
            self._add(key, position='left')
            self._add(value)
            data_values.append(dict({
                "key_id": id(key),
                "val_id": id(value),
            }))

        type_catagory, type_name = type_info
        self.data.append(dict({
            "id": obj_id,
            "type": type_name,
            "catagory": type_catagory,
            "options": {
                "inline_keys": not self.options.nolies,
                "inline_vals": self.options.inline,
                "position": kwargs.get('position', 'center'),
                "ellipsis": max_reached,
            },
            "values": data_values,
        }))

    def _add(self, obj, **kwargs):
        type_info = get_type_info(obj)
        if type_info[0] == 'key-value':
            self._add_key_value_object(obj, type_info, **kwargs)
        elif type_info[0] == 'sequence':
            self._add_sequence_object(obj, type_info, **kwargs)
        elif type_info[0] == 'array':
            self._add_array_object(obj, type_info, **kwargs)
        elif type_info[0] == 'class' or type_info[0].endswith('instance'):
            self._add_class_object(obj, type_info, **kwargs)
        else:
            self._add_object(obj, type_info, **kwargs)

    def add(self, filtered_locals):
        self._new_ids = list()
        for obj in filtered_locals.values():
            self._add(obj)

    def is_empty(self):
        return len(self) == 0

    def dumps(self):
        return json.dumps(self.data)

    def clean(self):
        return jsonutil.json_clean(self.data)

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)


class TraceHistory(object):

    def __init__(self, options):
        self.options = options
        self.stack_history = list()
        self.heap_history = list()
        self.output_history = list()

    def clear(self):
        self.__init__(self.options)

    def append(self, stack_frames, heap_objects, output):
        self.stack_history.append(stack_frames)
        self.heap_history.append(heap_objects)
        self.output_history.append(output)

    def sort_frame_locals(self):
        # Collect all unique object id's
        ids_sort_order = list()
        for heap in self.heap_history:
            for obj in heap:
                if obj['id'] not in ids_sort_order:
                    ids_sort_order.append(obj['id'])

        def id_sort_key(x):
            # Sort according to the order of object creation in the heap
            return ids_sort_order.index(x['id'])

        for stackframes in self.stack_history:
            for frame in stackframes:
                frame['vars'].sort(key=id_sort_key)

        for heap in self.heap_history:
            heap.data.sort(key=id_sort_key)

    def is_empty(self):
        return len(self) == 0

    def _to_dict(self):
        return dict({
            "stack_history": [x.data for x in self.stack_history],
            "heap_history": [x.data for x in self.heap_history],
            "output_history": self.output_history,
        })

    def dumps(self):
        return json.dumps(self._to_dict())

    def clean(self):
        return jsonutil.json_clean(self._to_dict())

    def __len__(self):
        return len(self.stack_history)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

from ipykernel.comm import Comm
from IPython.core import magic_arguments
from IPython.core.magic import Magics, cell_magic, magics_class
from IPython.core.magics.namespace import NamespaceMagics

from .debugger import Bdb


@magics_class
class NbtutorMagics(Magics):

    def __init__(self, shell):
        super(NbtutorMagics, self).__init__(shell)
        self.comm = Comm(target_name='nbtutor_comm')

    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        '-r', '--reset', action='store_true', default=False,
        help="Reset (clear) the IPython global namespace."
    )
    @magic_arguments.argument(
        '-f', '--force', action='store_true', default=False,
        help="Suppress the reset confirmation message."
    )
    @magic_arguments.argument(
        '-i', '--inline', action='store_true', default=False,
        help="Inline primitive objects."
    )
    @magic_arguments.argument(
        '-d', '--depth', metavar='N', type=int, default=1,
        help="The stack frame visualization depth (default: 1)."
    )
    @magic_arguments.argument(
        '--digits', metavar='D', type=int, default=3,
        help="The number of significant digits for floats (default: 3)."
    )
    @magic_arguments.argument(
        '--max_size', metavar='S', type=int, default=5,
        help="The number of sequence object elements to visualize (default: 5)."
    )
    @magic_arguments.argument(
        '--step_all', action='store_true', default=False,
        help="Step through all frames (including other global scope frames)"
    )
    @magic_arguments.argument(
        '--expand_arrays', action='store_true', default=False,
        help="Expand numpy arrays to show underlying data"
    )
    @magic_arguments.argument(
        '--nolies', action='store_true', default=False,
        help="No inlined keys, attributes or primitive objects"
    )
    @magic_arguments.argument(
        '--debug', action='store_true', default=False,
        help="Debug nbtutor."
    )
    @cell_magic
    def nbtutor(self, line, cell):
        opts = magic_arguments.parse_argstring(self.nbtutor, line)
        if opts.reset:
            params = '-f' if opts.force else ''
            NamespaceMagics(self.shell).reset(params)
        if opts.nolies:
            opts.inline = False

        bdb = Bdb(self.shell, opts)
        bdb.run_cell(cell)

        self.shell.run_cell(cell)
        # FIXME: This pointless re-running the cell again via ipython is needed
        # to get the "<ipython-input-{0}-{1}>" f_code.co_filename set and the
        # code cached. I don't know enough about the IPython API to do this
        # better inside the debugger.

        if not bdb.code_error or opts.debug:
            self.comm.send(bdb.trace_history.clean())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copied and modified from
# https://github.com/Anaconda-Platform/nbpresent/tree/master/nbpresent/tasks
from __future__ import absolute_import

import os
import sys

from subprocess import Popen

from tools import _env


def main(**opts):
    args = [
        _env.node_bin("lessc"),
        "--autoprefix",
        "--clean-css",
        "--include-path={}".format(_env.node_bin("..")),
        os.path.join(_env.SRC, "less", "index.less"),
        os.path.join(_env.DIST, "css", "nbtutor.min.css")
    ] + opts.get("less", [])
    return Popen(args, shell=_env.IS_WIN).wait()


if __name__ == "__main__":
    sys.exit(main())
